/*
 * p_main.c - The Python interface
 *
 * Copyright 2006   firkanim@mail.datanet.hu (Fazekas Laszlo)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* based on the xxmodule.c sample of the Python source */

#include <Python.h>

#include "l_call.h"

/* some declarations to be compatible with older Pythons */

#ifndef PyDoc_STR
#	define PyDoc_STR(str) str
#endif

#ifndef PyMODINIT_FUNC
#	if defined(__cplusplus)
#		define PyMODINIT_FUNC extern "C" void
#	else /* __cplusplus */
#		define PyMODINIT_FUNC void
#	endif /* !__cplusplus */
#endif

static PyObject *ErrorObject;

typedef struct {
	PyObject_HEAD
	PyObject	*x_attr;	/* Attributes dictionary */
	LCall_Struc	 lcs;		/* Lua interface */
} LipObject;

static PyTypeObject LipO_Type;

#define LipoObject_Check(v)	((v)->ob_type == &LipO_Type)

/* ------------------- the implemented functions --------------- */

static PyObject *lipo_dostring(PyObject *self, PyObject *args) {
	const char *cmd;

	if (!PyArg_ParseTuple(args, "s", &cmd)) return NULL;

	return Py_BuildValue("i", lcall_dostring(&(((LipObject *)self)->lcs), cmd));
}

/* ---------------- the administration for Python -------------- */

static LipObject *newLipObject(PyObject *arg)
{
	LipObject *self= PyObject_New(LipObject, &LipO_Type);

	if (self == NULL) return NULL;

	self->x_attr= NULL;
	lcall_init(&(self->lcs));
	return self;
}

/* Lipo methods */

static void Lipo_dealloc(LipObject *self) {
	Py_XDECREF(self->x_attr);
	lcall_exit(&(self->lcs));
	PyObject_Del(self);
}

static PyMethodDef Lipo_methods[] = {
	{ "dostring",
	  (PyCFunction)lipo_dostring, METH_VARARGS,
	  PyDoc_STR("execute a lua string") },

	{ NULL, NULL, 0, NULL }		/* sentinel */
};

static PyObject *Lipo_getattr(LipObject *self, char *name)
{
	if (self->x_attr != NULL) {
		PyObject *v = PyDict_GetItemString(self->x_attr, name);
		if (v != NULL) {
			Py_INCREF(v);
			return v;
		}
	}
	return Py_FindMethod(Lipo_methods, (PyObject *)self, name);
}

static int Lipo_setattr(LipObject *self, char *name, PyObject *v)
{
	if (self->x_attr == NULL) {
		self->x_attr = PyDict_New();

		if (self->x_attr == NULL) return -1;
	}

	if (v == NULL) {
		int rv = PyDict_DelItemString(self->x_attr, name);
		if (rv < 0)
			PyErr_SetString(PyExc_AttributeError,
			  "delete non-existing Lipo attribute");

		return rv;

	} else return PyDict_SetItemString(self->x_attr, name, v);
}

static PyTypeObject LipO_Type= {

	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,				/*ob_size*/
	"_Lipmodule._Lipo",		/*tp_name*/
	sizeof(LipObject),		/*tp_basicsize*/
	0,				/*tp_itemsize*/
	/* methods */
	(destructor)Lipo_dealloc,	/*tp_dealloc*/
	0,				/*tp_print*/
	(getattrfunc)Lipo_getattr,	/*tp_getattr*/
	(setattrfunc)Lipo_setattr,	/*tp_setattr*/
	0,				/*tp_compare*/
	0,				/*tp_repr*/
	0,				/*tp_as_number*/
	0,				/*tp_as_sequence*/
	0,				/*tp_as_mapping*/
	0,				/*tp_hash*/
        0,				/*tp_call*/
        0,				/*tp_str*/
        0,				/*tp_getattro*/
        0,				/*tp_setattro*/
        0,				/*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,		/*tp_flags*/
        0,				/*tp_doc*/
        0,				/*tp_traverse*/
        0,				/*tp_clear*/
        0,				/*tp_richcompare*/
        0,				/*tp_weaklistoffset*/
        0,				/*tp_iter*/
        0,				/*tp_iternext*/
        0,				/*tp_methods*/
        0,				/*tp_members*/
        0,				/*tp_getset*/
        0,				/*tp_base*/
        0,				/*tp_dict*/
        0,				/*tp_descr_get*/
        0,				/*tp_descr_set*/
        0,				/*tp_dictoffset*/
        0,				/*tp_init*/
        0,				/*tp_alloc*/
        0,				/*tp_new*/
        0,				/*tp_free*/
        0				/*tp_is_gc*/
};

static PyObject *lip_new(PyObject *self, PyObject *args)
{
	LipObject *rv;

	if (!PyArg_ParseTuple(args, ":new")) return NULL;

	rv= newLipObject(args);

	if (rv == NULL) return NULL;

	return (PyObject *)rv;
}

/* List of functions defined in the module */

static PyMethodDef lip_methods[] = {
	{ "new", lip_new, METH_VARARGS, PyDoc_STR("new() -> new _Lip object") },

	{ NULL, NULL, 0, NULL }		/* sentinel */
};

/* Initialization function, it must be called 'init_lip'! */

PyMODINIT_FUNC init_lip(void) {
	PyObject *m;

	/* Finalize the type object including setting type of the new type
	 * object; doing it here is required for portability to Windows 
	 * without requiring C++. */

	if (PyType_Ready(&LipO_Type) < 0) return;

	/* Create the module and add the functions */

	m= Py_InitModule3("_lip", lip_methods, PyDoc_STR("_lip: Lua in Python"));

	if (m == NULL) return;

	if (ErrorObject == NULL) {
		ErrorObject = PyErr_NewException("_lip.error", NULL, NULL);

		if (ErrorObject == NULL) return;
	}
	Py_INCREF(ErrorObject);
	PyModule_AddObject(m, "error", ErrorObject);
}
